/*
 * HP XLFD Enhancements Utilities
 *
 * This file supports the HP XLFD Enhancements to the XLFD name, and is
 * based on the font server's fontxlfd.c file by Keith Packard of the
 * MIT X Consortium.
 *
 * Author:  Nathan Meyers, Hewlett-Packard Company
 *
 * $Header: xlfdutil.c,v 1.9 92/07/02 16:25:38 nathanm Exp $
 */

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "xlfdutil.h"

#ifndef M_PI
#define M_PI 3.1415926535897932384
#endif

/*
 * This module contains two tools for taking advantage of the HP XLFD
 * enhancements:  ParseXLFD() and XDrawRotatedString().  This comment
 * block describes these two tools in detail.
 *
 *  === ParseXLFD(char *Fontname, XLFDRecPtr Record, int Option) ===
 *
 * ParseXLFD() is a tool for disassembling and reassembling XLFD font
 * names, while allowing for easy modifications of fields and subfields
 * within the XLFD name, and accommodating the HP XLFD Enhancements.
 *
 * ParseXLFD() is based on the Font Server's FontParseXLFDName()
 * procedure, with changes to make it more useful on the client side.
 * Certain aspects of ParseXLFD()'s design (in particular, its memory
 * management) might not be ideal for all applications, in which case we
 * encourage the developer to modify this source to meet his needs.
 *
 * An explanation of ParseXLFD()'s four modes of behavior:
 *
 *   1) When called with Option = FONT_XLFD_REPLACE_NONE, ParseXLFD()
 *	will parse the Fontname parameter, and fill in the fields of the
 *	Record parameter with values representing the various fields and
 *	subfields in the name.  All char* fields point to malloc'd text
 *	strings that contain copies of the relevant fields from the
 *	original name.
 *
 *	For example, when called with a Fontname of:
 *
 *	  -adobe-courier-medium-r-normal--12-120-75-75-m-30-iso8859-1
 *
 *	ParseXLFD will fill in the fields of Record as follows:
 *
 *	  foundry = "adobe"
 *	  familyName = "courier"
 *	  weightName = "medium"
 *	  slant = "r"
 *	  setwidthName = "normal"
 *	  pixelSize = 12
 *	  pointSize = 120
 *	  resolutionX = 75
 *	  resolutionY = 75
 *	  spacing = "m"
 *	  averageWidth = 30
 *	  charsetRegistry = "iso8859"
 *	  charsetEncoding = "1"
 *	  pixelSetsize = 12
 *	  pointSetsize = 120
 *	  horizWeight = 0
 *	  vertWeight = 0
 *	  slantAngle = 0
 *	  mirrorX = False
 *	  mirrorY = False
 *	  rotationAngle = 0
 *	  ranges = ""
 *
 *	If HP XLFD Enhancements are utilized, the enhancement fields
 *	take on more interesting values.  For example:
 *
 *	  -adobe-courier-medium-r_1000-normal-+500+mx-0-120+240-75-75-m-0-iso8859-1=65:70
 *
 *	fills in the following fields:
 *
 *	  foundry = "adobe"
 *	  familyName = "courier"
 *	  weightName = "medium"
 *	  slant = "r"
 *	  setwidthName = "normal"
 *	  pixelSize = 0
 *	  pointSize = 120
 *	  resolutionX = 75
 *	  resolutionY = 75
 *	  spacing = "m"
 *	  averageWidth = 0
 *	  charsetRegistry = "iso8859"
 *	  charsetEncoding = "1"
 *	  pixelSetsize = 0
 *	  pointSetsize = 240
 *	  horizWeight = 0
 *	  vertWeight = 0
 *	  slantAngle = -1000
 *	  mirrorX = True
 *	  mirrorY = False
 *	  rotationAngle = 500
 *	  ranges = "65:70"
 *
 *   2) When called with Option = FONT_XLFD_REPLACE_STAR, ParseXLFD()
 *	parses the name as above, filling in Record as above, and also
 *	computes a new version of the Fontname in which the enhancement
 *	fields are removed and the scalable fields are replaced with the
 *	"*" wildcard character.  This new version of the name is a
 *	malloc'd string returned as the function's return value.  (Note
 *	that in the code on which this is based, the new name overwrote
 *	the data pointed to by the Fontname parameter).
 *
 *   3) When called with Option = FONT_XLFD_REPLACE_ZERO, ParseXLFD()
 *	parses the name as above, filling in Record as above, and also
 *	computes a new version of the Fontname in which the enhancement
 *	fields are removed and the scalable fields are replaced with the
 *	"0" character.  This new version of the name is a malloc'd
 *	string returned as the function's return value.  (Note that in
 *	the code on which this is based, the new name overwrote the data
 *	pointed to by the Fontname parameter).
 *
 *   4) When called with Option = FONT_XLFD_REPLACE_VALUE, ParseXLFD()
 *	creates an XLFD string out of the data contained in the Record
 *	structure, returning the malloc'd name as the return value.  The
 *	Fontname argument is ignored (in the original code on which this
 *	is based, Fontname was used as a pointer to where the string was
 *	to be written).
 *
 * For all Option values, a return value of NULL indicates the inability
 * of ParseXLFD() to do its job:  usually a bad fontname or malloc()
 * failure.
 *
 *
 *
 * 
 *  === XDrawRotatedString(Display *display, Drawable d, GC gc, int x,
 *			   int y, char *string, int length,
 *			   XFontStruct *font_struct, int angle) ===
 *
 * XDrawRotatedString() draws text at an angle.  The first seven
 * parameters are identical to those for XDrawString() -- the last two
 * provide the additional information needed to render the text:  the
 * angle at which the text will be drawn (specified in 1/64ths of a
 * degree) and the font information structure containing the font's
 * metrics.
 *
 * Using the font metrics from font_struct, and the given angle,
 * XDrawRotatedString() performs the necessary computations and places
 * the characters correctly for non-horizontal rendering.
 *
 * The XDrawRotatedString() interface was designed to be similar to
 * XDrawString()'s.  While this makes it easy to use, the result is not
 * the fastest possible implementation -- developers may wish to modify
 * the code given below for faster operation.  In particular, this
 * implementation performs two trigonometric computations and retrieves
 * two atoms -- in order to compute sintheta and costheta -- every time
 * it is called.  If the same font is to be used for more than one
 * XDrawRotatedString() call, a more efficient implementation would be
 * to compute and keep the sintheta and costheta values in the calling
 * code and pass them to this procedure.
 *
 *
 *
 *  === XDrawRotatedImageString() ===
 *
 * Why isn't there a XDrawRotatedImageString() procedure?
 *
 * Well, if there were, it would work by computing the width of the
 * entire string [width = XTextWidth(font_struct, string, length);] and
 * then drawing a filled polygon in the background color whose vertices
 * were (give or take rounding to the nearest pixel):
 *
 *	XPoint points[4];
 *
 *	points[0].x = -font_struct->descent * sintheta + x;
 *	points[0].y = font_struct->descent * costheta + y;
 *
 *	points[1].x = -font_struct->descent * sintheta + width * costheta + x;
 *	points[1].y = font_struct->descent * costheta + width * sintheta + y;
 *
 *	points[2].x = font_struct->ascent * sintheta + width * costheta + x;
 *	points[2].y = -font_struct->ascent * costheta + width * sintheta + y;
 *
 *	points[3].x = font_struct->ascent * sintheta + x;
 *	points[3].y = -font_struct->ascent * costheta + y;
 *
 * It would then call XDrawRotatedString() to draw the text.
 *
 * It would if it could.  But X won't draw a polygon in the background
 * color and won't let us query the GC to find out what the background
 * color is.  So, for clients that need image-text-like behavior, the
 * above information should help them figure out how to draw the
 * background.
 *
 */

static int
#ifdef __STDC__
dashcount(char *Fontname)
#else /* __STDC__ */
dashcount(Fontname)
char *Fontname;
#endif /* __STDC__ */
{
    register char *p;
    int count = 0;

    if (Fontname == NULL) return 0;

    for (p = Fontname; *p; p++) if (*p == '-') count++;

    return count;
}

static void
#ifdef __STDC__
free_record_fields(XLFDRecPtr Record)
#else /* __STDC__ */
free_record_fields(Record)
XLFDRecPtr Record;
#endif /* __STDC__ */
{
    if (Record->foundry) free(Record->foundry);
    if (Record->familyName) free(Record->familyName);
    if (Record->weightName) free(Record->weightName);
    if (Record->slant) free(Record->slant);
    if (Record->setwidthName) free(Record->setwidthName);
    if (Record->addStyleName) free(Record->addStyleName);
    if (Record->spacing) free(Record->spacing);
    if (Record->charsetRegistry) free(Record->charsetRegistry);
    if (Record->charsetEncoding) free(Record->charsetEncoding);
    if (Record->ranges) free(Record->ranges);
    Record->foundry =
    Record->familyName =
    Record->weightName =
    Record->slant =
    Record->setwidthName =
    Record->addStyleName =
    Record->spacing =
    Record->charsetRegistry =
    Record->charsetEncoding =
    Record->ranges = NULL;
}


static char *
#ifdef __STDC__
getXLFDfield(register char *namein, char **nameout)
#else /* __STDC__ */
getXLFDfield(namein, nameout)
register char *namein;
char **nameout;
#endif /* __STDC__ */
{
    register char *p1, *p2;
    char *result;

    if (namein == NULL) return NULL;

    /* Find the next hyphen */
    for (p2 = namein; *p2 && *p2 != '-'; p2++);
    if (*p2 != '-') return *nameout = NULL;

    /* Make space for this field */
    if ((result = (char *)malloc(p2 - namein + 1)) == NULL)
	return *nameout = NULL;

    /* Copy this field */
    for (p1 = result; namein != p2; p1++, namein++) *p1 = *namein;
    *p1 = '\0';

    /* Update *nameout */
    *nameout = p2 + 1;

    /* Done */
    return result;
}

static char *
#ifdef __STDC__
getXLFDweightField(char *namein, int *hw, int *vw, char **nameout)
#else /* __STDC__ */
getXLFDweightField(namein, hw, vw, nameout)
char *namein;
int *hw, *vw;
char **nameout;
#endif /* __STDC__ */
{
    char *p1;
    register char *p2;
    char *result;
    int negative;
    long strtol();

    if (namein == NULL) return NULL;

    /* Find the next separator */
    for (p2 = namein; *p2 && *p2 != '-' && *p2 != '+' && *p2 != '_'; p2++);
    if (*p2 == '\0') return *nameout = NULL;

    /* Make space for this field */
    if ((result = (char *)malloc(p2 - namein + 1)) == NULL)
	return *nameout = NULL;

    /* Copy this field */
    for (p1 = result; namein != p2; p1++, namein++) *p1 = *namein;
    *p1 = '\0';

    /* Now to the numbers.  First, initialize them. */
    *hw = *vw = 0;

    /* Are there any numbers? */
    if ((negative = *p2 == '_') || *p2 == '+')
    {
	/* Yes, get the first number */
	int number = strtol(++p2, &p1, 10);
	if (*p2 == '-' || p1 == p2) return free(result), *nameout = NULL;
	*hw = *vw = negative ? -number : number;
	p2 = p1;

	/* Is there a second number? */
	if ((negative = *p2 == '_') || *p2 == '+')
	{
	    /* Yes... get it. */
	    number = strtol(++p2, &p1, 10);
	    if (*p2 == '-' || p1 == p2) return free(result), *nameout = NULL;
	    *vw = negative ? -number : number;
	    p2 = p1;
	}
    }

    if (*p2 != '-') return free(result), *nameout = NULL;

    /* Update *nameout */
    *nameout = p2 + 1;

    /* Done */
    return result;
}

static char *
#ifdef __STDC__
getXLFDslantField(char *namein, int *slant, char **nameout)
#else /* __STDC__ */
getXLFDslantField(namein, slant, nameout)
char *namein;
int *slant;
char **nameout;
#endif /* __STDC__ */
{
    char *p1;
    register char *p2;
    char *result;
    int negative;
    long strtol();

    if (namein == NULL) return NULL;

    /* Find the next separator */
    for (p2 = namein; *p2 && *p2 != '-' && *p2 != '+' && *p2 != '_'; p2++);
    if (*p2 == '\0') return *nameout = NULL;

    /* Make space for this field */
    if ((result = (char *)malloc(p2 - namein + 1)) == NULL)
	return *nameout = NULL;

    /* Copy this field */
    for (p1 = result; namein != p2; p1++, namein++) *p1 = *namein;
    *p1 = '\0';

    /* Now to the slant number.  First, initialize it. */
    *slant = 0;

    /* Is there a number? */
    if ((negative = *p2 == '_') || *p2 == '+')
    {
	/* Yes, get it */
	int number = strtol(++p2, &p1, 10);
	if (*p2 == '-' || p1 == p2) return free(result), *nameout = NULL;
	*slant = negative ? -number : number;
	p2 = p1;
    }

    if (*p2 != '-') return free(result), *nameout = NULL;

    /* Update *nameout */
    *nameout = p2 + 1;

    /* Done */
    return result;
}

static int
#ifdef __STDC__
strnicmp(char *s1, char *s2, int count)
#else /* __STDC__ */
strnicmp(s1, s2, count)
char *s1;
char *s2;
int count;
#endif /* __STDC__ */
{
    int diff;

    while (count-- > 0)
	if (diff = tolower(*s1++) - tolower(*s2++))
	    return diff;

    return 0;
}

static char *
#ifdef __STDC__
getXLFDaddStyleField(char *namein, int *mx, int *my, int *angle, char **nameout)
#else /* __STDC__ */
getXLFDaddStyleField(namein, mx, my, angle, nameout)
char *namein;
int *mx, *my, *angle;
char **nameout;
#endif /* __STDC__ */
{
    char *p1;
    register char *p2;
    char *result;
    int negative;
    long strtol();

    if (namein == NULL) return NULL;

    /* Find the next separator */
    for (p2 = namein; *p2 && *p2 != '-' && *p2 != '+' && *p2 != '_'; p2++);
    if (*p2 == '\0') return *nameout = NULL;

    /* Make space for this field */
    if ((result = (char *)malloc(p2 - namein + 1)) == NULL)
	return *nameout = NULL;

    /* Copy this field */
    for (p1 = result; namein != p2; p1++, namein++) *p1 = *namein;
    *p1 = '\0';

    /* Now to the enhancements.  First, initialize them. */
    *mx = *my = False;
    *angle = 0;

    /* Any enhancements? */
    while ((negative = *p2 == '_') || *p2 == '+')
    {
	int number;

	/* Yes, get the next enhancement */
	if (!strnicmp(p2, "+mx", 3))
	{
	    *mx = True;
	    p2 += 3;
	    continue;
	}
	if (!strnicmp(p2, "+my", 3))
	{
	    *my = True;
	    p2 += 3;
	    continue;
	}
	number = strtol(++p2, &p1, 10);
	if (*p2 == '-' || p1 == p2) return free(result), *nameout = NULL;
	*angle = negative ? -number : number;
	p2 = p1;
    }

    if (*p2 != '-') return free(result), *nameout = NULL;

    /* Update *nameout */
    *nameout = p2 + 1;

    /* Done */
    return result;
}

static void
#ifdef __STDC__
getXLFDtwoNumbers(char *namein, int *n1, int *n2, char **nameout)
#else /* __STDC__ */
getXLFDtwoNumbers(namein, n1, n2, nameout)
char *namein;
int *n1, *n2;
char **nameout;
#endif /* __STDC__ */
{
    char *p1;
    int number;
    long strtol();

    if (namein == NULL) return;

    /* Initialize our numbers */
    *n1 = *n2 = 0;

    /* Get the first number */
    number = strtol(namein, &p1, 10);
    if (*namein == '-' || p1 == namein)
    {
	*nameout = NULL;
	return;
    }
    *n1 = *n2 = number;
    namein = p1;

    /* Is there a second number? */
    if (*namein == '+')
    {
	/* Yes... get it. */
	number = strtol(++namein, &p1, 10);
	if (*namein == '-' || p1 == namein)
	{
	    *nameout = NULL;
	    return;
	}
	*n2 = number;
	namein = p1;
    }

    if (*namein != '-')
    {
	*nameout = NULL;
	return;
    }

    /* Update *nameout */
    *nameout = namein + 1;

    /* Done */
}

static char *
#ifdef __STDC__
getXLFDnumber(char *namein, int *n1, char **nameout)
#else /* __STDC__ */
getXLFDnumber(namein, n1, nameout)
char *namein;
int *n1;
char **nameout;
#endif /* __STDC__ */
{
    char *p1;
    int number;
    long strtol();

    if (namein == NULL) return;

    /* Initialize our numbers */
    *n1 = 0;

    /* Get the first number */
    number = strtol(namein, &p1, 10);
    if (*namein == '-' || p1 == namein)
    {
	*nameout = NULL;
	return;
    }
    *n1 = number;
    namein = p1;

    if (*namein != '-')
    {
	*nameout = NULL;
	return;
    }

    /* Update *nameout */
    *nameout = namein + 1;

    /* Done */
}

static char *
#ifdef __STDC__
getXLFDencodingField(char *namein, char **nameout)
#else /* __STDC__ */
getXLFDencodingField(namein, nameout)
char *namein;
char **nameout;
#endif /* __STDC__ */
{
    register char *p1, *p2;
    char *result;

    if (namein == NULL) return NULL;

    /* Find the next separator */
    for (p2 = namein; *p2 && *p2 != '='; p2++);
    if (*p2 && *p2 != '=') return *nameout = NULL;

    /* Make space for this field */
    if ((result = (char *)malloc(p2 - namein + 1)) == NULL)
	return *nameout = NULL;

    /* Copy this field */
    for (p1 = result; namein != p2; p1++, namein++) *p1 = *namein;
    *p1 = '\0';

    /* Update *nameout */
    *nameout = p2;

    /* Done */
    return result;
}

static char *
#ifdef __STDC__
getXLFDrangesField(char *namein, char **nameout)
#else /* __STDC__ */
getXLFDrangesField(namein, nameout)
char *namein;
char **nameout;
#endif /* __STDC__ */
{
    register char *p1, *p2;
    char *result;

    /* This procedure copies the XLFD ranges field without checking
       whether it looks valid */

    if (namein == NULL) return NULL;
    if (*namein == '=') namein++;
    else if (*namein != '\0') return *nameout = NULL;

    /* Find the end of the line */
    for (p2 = namein; *p2; p2++);

    /* Make space for this field */
    if ((result = (char *)malloc(p2 - namein + 1)) == NULL)
	return *nameout = NULL;

    /* Copy this field */
    for (p1 = result; namein != p2; p1++, namein++) *p1 = *namein;
    *p1 = '\0';

    /* Update *nameout */
    *nameout = p2;

    /* Done */
    return result;
}

static char *
#ifdef __STDC__
format_weight(int hw, int vw)
#else /* __STDC__ */
format_weight(hw, vw)
int hw;
int vw;
#endif /* __STDC__ */
{
    static char buffer[100];

    if (hw || vw)
    {
	int s1 = '+';
	int s2 = '+';
	if (hw < 0)
	{
	    s1 = '_';
	    hw = -hw;
	}
	if (vw < 0)
	{
	    s2 = '_';
	    vw = -vw;
	}
	if (vw == hw) sprintf(buffer, "%c%d", s1, hw);
	else sprintf(buffer, "%c%d%c%d", s1, hw, s2, vw);
    }
    else
	buffer[0] = '\0';

    return buffer;
}

static char *
#ifdef __STDC__
format_slant(int slant)
#else /* __STDC__ */
format_slant(slant)
int slant;
#endif /* __STDC__ */
{
    static char buffer[100];

    if (slant)
    {
	int s1 = '+';
	if (slant < 0)
	{
	    s1 = '_';
	    slant = -slant;
	}
	sprintf(buffer, "%c%d", s1, slant);
    }
    else
	buffer[0] = '\0';

    return buffer;
}

static char *
#ifdef __STDC__
format_addstyle(Bool mx, Bool my, int angle)
#else /* __STDC__ */
format_addstyle(mx, my, angle)
Bool mx;
Bool my;
int angle;
#endif /* __STDC__ */
{
    static char buffer[100];

    buffer[0] = '\0';
    if (mx) strcat(buffer, "+mx");
    if (my) strcat(buffer, "+my");
    if (angle)
    {
	int s1 = '+';
	if (angle < 0)
	{
	    s1 = '_';
	    angle = -angle;
	}
	sprintf(buffer + strlen(buffer), "%c%d", s1, angle);
    }
    return buffer;
}

static char *
#ifdef __STDC__
format_2_num(int n1, int n2)
#else /* __STDC__ */
format_2_num(n1, n2)
int n1;
int n2;
#endif /* __STDC__ */
{
    static char buffer[100];

    if (n1 < 0 || n2 < 0) strcpy(buffer, "*");
    else if (n1 == n2) sprintf(buffer, "%d", n1);
    else sprintf(buffer, "%d+%d", n1, n2);

    return buffer;
}

static char *
#ifdef __STDC__
format_1_num(int n1)
#else /* __STDC__ */
format_1_num(n1)
int n1;
#endif /* __STDC__ */
{
    static char buffer[100];

    if (n1 < 0) strcpy(buffer, "*");
    else sprintf(buffer, "%d", n1);

    return buffer;
}

char *
#ifdef __STDC__
ParseXLFD(char *Fontname, XLFDRecPtr Record, int Option)
#else /* __STDC__ */
ParseXLFD(Fontname, Record, Option)
char *Fontname;
XLFDRecPtr Record;
int Option;
#endif /* __STDC__ */
{
    XLFDRec temprec;
    int free_temprec = 0;
    char *result;
    int buffersize;

    if (Option != FONT_XLFD_REPLACE_VALUE &&
	dashcount(Fontname) != 14) return NULL;

    /* If we are called with one of the parse options */
    if (Option == FONT_XLFD_REPLACE_NONE ||
	Option == FONT_XLFD_REPLACE_STAR ||
	Option == FONT_XLFD_REPLACE_ZERO)
    {
	/* If no Record specified, we'll use a temporary one */
	if (Record == NULL)
	{
	    Record = &temprec;
	    free_temprec = 1;
	}

	/* Parse the fields from Fontname */

	while (*Fontname && *Fontname++ != '-');

	Record->foundry = getXLFDfield(Fontname, &Fontname);
	Record->familyName = getXLFDfield(Fontname, &Fontname);
	Record->weightName = getXLFDweightField(Fontname,
						&Record->horizWeight,
						&Record->vertWeight,
						&Fontname);
	Record->slant = getXLFDslantField(Fontname,
					  &Record->slantAngle,
					  &Fontname);
	Record->setwidthName = getXLFDfield(Fontname, &Fontname);
	Record->addStyleName = getXLFDaddStyleField(Fontname,
						    &Record->mirrorX,
						    &Record->mirrorY,
						    &Record->rotationAngle,
						    &Fontname);
	getXLFDtwoNumbers(Fontname,
			  &Record->pixelSize,
			  &Record->pixelSetsize,
			  &Fontname);
	getXLFDtwoNumbers(Fontname,
			  &Record->pointSize,
			  &Record->pointSetsize,
			  &Fontname);
	getXLFDnumber(Fontname,
		      &Record->resolutionX,
		      &Fontname);
	getXLFDnumber(Fontname,
		      &Record->resolutionY,
		      &Fontname);
	Record->spacing = getXLFDfield(Fontname, &Fontname);
	getXLFDnumber(Fontname,
		      &Record->averageWidth,
		      &Fontname);
	Record->charsetRegistry = getXLFDfield(Fontname, &Fontname);
	Record->charsetEncoding = getXLFDencodingField(Fontname, &Fontname);
	Record->ranges = getXLFDrangesField(Fontname, &Fontname);

	/* Did everything work? */
	if (Fontname == NULL)
	{
	    /* No.  Clean up and return. */
	    free_record_fields(Record);
	    return NULL;
	}

	/* Prepare for building of the new name */
	if (Option != FONT_XLFD_REPLACE_NONE)
	{
	    if (Record != &temprec)
	    {
		temprec = *Record;
		Record = &temprec;
	    }

	    temprec.resolutionX = temprec.resolutionY = -1;
	    temprec.horizWeight = 0;
	    temprec.vertWeight = 0;
	    temprec.slantAngle = 0;
	    temprec.mirrorX = False;
	    temprec.mirrorY = False;
	    temprec.rotationAngle = 0;
	    temprec.ranges = NULL;

	    if (Option == FONT_XLFD_REPLACE_STAR)
		temprec.pixelSize =
		temprec.pixelSetsize =
		temprec.pointSize =
		temprec.pointSetsize = -1;
	    else
		temprec.pixelSize =
		temprec.pixelSetsize =
		temprec.pointSize =
		temprec.pointSetsize = 0;
	}
    }

    if (Option == FONT_XLFD_REPLACE_NONE)
    {
	if (free_temprec) free_record_fields(&temprec);
	return Fontname;
    }

#define TEXTOK(s) ((s) ? (s) : "")

    /* Time to put together a new name.  First precompute its length */
    buffersize = strlen(TEXTOK(Record->foundry)) +
		 strlen(TEXTOK(Record->familyName)) +
		 strlen(TEXTOK(Record->weightName)) +
		 strlen(format_weight(Record->horizWeight,
				      Record->vertWeight)) +
		 strlen(TEXTOK(Record->slant)) +
		 strlen(format_slant(Record->slantAngle)) +
		 strlen(TEXTOK(Record->setwidthName)) +
		 strlen(TEXTOK(Record->addStyleName)) +
		 strlen(format_addstyle(Record->mirrorX,
					Record->mirrorY,
					Record->rotationAngle)) +
		 strlen(format_2_num(Record->pixelSize,
				     Record->pixelSetsize)) +
		 strlen(format_2_num(Record->pointSize,
				     Record->pointSetsize)) +
		 strlen(format_1_num(Record->resolutionX)) +
		 strlen(format_1_num(Record->resolutionY)) +
		 strlen(TEXTOK(Record->spacing)) +
		 strlen(format_1_num(Record->averageWidth)) +
		 strlen(TEXTOK(Record->charsetRegistry)) +
		 strlen(TEXTOK(Record->charsetEncoding)) +
		 strlen(TEXTOK(Record->ranges)) +
		 16;

    if ((result = (char *)malloc(buffersize)) == NULL)
    {
	if (Option != FONT_XLFD_REPLACE_VALUE) free_record_fields(Record);
	return NULL;
    }

    strcpy(result, "-");
    strcat(result,TEXTOK(Record->foundry));
    strcat(result, "-");
    strcat(result,TEXTOK(Record->familyName));
    strcat(result, "-");
    strcat(result,TEXTOK(Record->weightName));
    strcat(result,format_weight(Record->horizWeight,
				Record->vertWeight));
    strcat(result, "-");
    strcat(result,TEXTOK(Record->slant));
    strcat(result,format_slant(Record->slantAngle));
    strcat(result, "-");
    strcat(result,TEXTOK(Record->setwidthName));
    strcat(result, "-");
    strcat(result,TEXTOK(Record->addStyleName));
    strcat(result,format_addstyle(Record->mirrorX,
				  Record->mirrorY,
				  Record->rotationAngle));
    strcat(result, "-");
    strcat(result,format_2_num(Record->pixelSize,
			       Record->pixelSetsize));
    strcat(result, "-");
    strcat(result,format_2_num(Record->pointSize,
			       Record->pointSetsize));
    strcat(result, "-");
    strcat(result,format_1_num(Record->resolutionX));
    strcat(result, "-");
    strcat(result,format_1_num(Record->resolutionY));
    strcat(result, "-");
    strcat(result,TEXTOK(Record->spacing));
    strcat(result, "-");
    strcat(result,format_1_num(Record->averageWidth));
    strcat(result, "-");
    strcat(result,TEXTOK(Record->charsetRegistry));
    strcat(result, "-");
    strcat(result,TEXTOK(Record->charsetEncoding));
    if (Record->ranges && *Record->ranges)
    {
	strcat(result, "=");
	strcat(result,TEXTOK(Record->ranges));
    }

    if (free_temprec) free_record_fields(&temprec);

    return result;
}

#define roundmult(x, y) (int)(((y) > 0) ? (x) * (y) + .5 : (x) * (y) - .5)

void
#ifdef __STDC__
XDrawRotatedString(Display *display, Drawable d, GC gc, int x, int y,
		   char *string, int length, XFontStruct *font_struct,
		   int angle)
#else /* __STDC__ */
XDrawRotatedString(display, d, gc, x, y, string, length, font_struct, angle)
Display *display;
Drawable d;
GC gc;
int x;
int y;
char *string;
int length;
XFontStruct *font_struct;
int angle;
#endif /* __STDC__ */
{
    double theta, sintheta, costheta, sin(), cos();
    int i, offset = 0, x_res, y_res;
    Atom atom;

    /* Compute X- and Y-components of our rendering angle */
    theta = (double)angle * M_PI / 11520.0;
    sintheta = -sin(theta);
    costheta = cos(theta);

    /* Compensate for non-square pixels */
    x_res = y_res = 100;
    atom = XInternAtom(display, "RESOLUTION_X", True);
    if (atom != None)
    for (i = 0; i < font_struct->n_properties; i++)
	if (font_struct->properties[i].name == atom)
	{
	    x_res = font_struct->properties[i].card32;
	    break;
	}
    atom = XInternAtom(display, "RESOLUTION_Y", True);
    if (atom != None)
    for (i = 0; i < font_struct->n_properties; i++)
	if (font_struct->properties[i].name == atom)
	{
	    y_res = font_struct->properties[i].card32;
	    break;
	}
    sintheta *= (double)y_res / (double)x_res;

    offset = 0;
    while (length-- > 0)
    {
	XDrawString(display, d, gc,
		    x + roundmult(offset, costheta),
		    y + roundmult(offset, sintheta),
		    string, 1);
	offset += XTextWidth(font_struct, string, 1);
	string++;
    }
}
